import logging
import logging.config
logger = logging.getLogger()
import os
import sys
from tinydb import  Query

def colons_to_slashes():
    def transform(doc):
        res = doc.get('i', None)
        if res:
            doc['i'] = doc['i'].replace(':', '/')
        return doc
    return transform

def main():
    import etm.data as data
    etmdir = os.getcwd()

    loglevel = 2 # info
    log_levels = [str(x) for x in range(1, 6)]
    if len(sys.argv) > 1 and sys.argv[1] in log_levels:
        loglevel = sys.argv.pop(1)
    if len(sys.argv) > 1 and os.path.isdir(sys.argv[1]):
        etmdir = sys.argv.pop(1)

    logdir = os.path.normpath(os.path.join(etmdir, 'logs'))
    if not os.path.isdir(logdir):
        os.makedirs(logdir)

    dbfile = os.path.normpath(os.path.join(etmdir, 'db.json'))
    ETMDB = data.initialize_tinydb(dbfile)
    DBITEM = ETMDB.table('items', cache_size=None)
    DBARCH = ETMDB.table('archive', cache_size=None)

    Reminder = Query()
    ETMDB.update(colons_to_slashes(), Reminder.i.exists())

if __name__ == "__main__":
    main()




import os

# import math
import sys
import shutil
import textwrap
import re
import csv
from dateutil.parser import parse as dateutil_parse
from dateutil.parser import parserinfo

from copy import deepcopy

# from etm.__main__ import setup_logging

from tinydb import where, Query

from prompt_toolkit import PromptSession

from etm.view import ETMQuery, format_week

from etm.model import format_datetime, parse, parse_duration, fmt_week

from etm.model import format_hours_and_tenths
import etm.data as data
from etm.common import (
    parse,
    ETM_CHAR,
    Period,
)
import itertools
from itertools import groupby

flatten = itertools.chain.from_iterable
from operator import itemgetter
from datetime import datetime, date, timedelta

ZERO = timedelta(seconds=0)
ONEMIN = timedelta(seconds=60)

finished_char = '\u2713'  #  ✓

minus_regex = re.compile(r'\s+\-(?=[a-zA-Z])')
groupdate_regex = re.compile(
    r'\bW{1,4}\b|\bY{1,4}\b|\b[M]{1,4}\b|\bd{3,4}\b|\b[D]{1,2}\b|\b[w]\b'
)


ETMDB = None
DBITEM = None
DBARCH = None

# pendulum to datetime format conversions
p2d = {
    'YY': '%y',
    'YYYY': '%Y',
    'M': '%-m',
    'MM': '%m',
    'MMM': '%b',
    'MMMM': '%B',
    'D': '%-d',
    'DD': '%d',
    'ddd': '%a',
    'dddd': '%A',
}


def daybeg():
    return datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)


def dayend():
    return datetime.now().replace(hour=23, minute=59, second=59, microsecond=0)


def weekbeg():
    d = daybeg()
    return d - timedelta(days=d.weekday())


def weekend():
    d = dayend()
    return d + timedelta(days=6 - d.weekday())


def monthbeg():
    d = daybeg()
    return d.replace(day=1)


def monthend():
    mb = monthbeg()
    (y, m) = (mb.year, mb.month + 1) if mb.month < 12 else (mb.year + 1, 1)
    return mb.replace(year=y, month=m) - timedelta(seconds=1)


def prevmonthend():
    return monthbeg() - timedelta(seconds=1)


def prevmonthbeg():
    me = prevmonthend()
    return me.replace(day=1)


date_shortcuts = {
    'daybeg': daybeg,
    'dayend': dayend,
    'weekbeg': weekbeg,
    'weekend': weekend,
    'monthbeg': monthbeg,
    'monthend': monthend,
    'prevmonthbeg': prevmonthbeg,
    'prevmonthend': prevmonthend,
}


def parse_reldt(s):
    """
    s takes the form datetime str [+-] duration str
    """
    msg = []
    plus = r'\s[+]\s'
    minus = r'\s[-]\s'
    sign = ''
    if re.search(plus, s):
        sign = '+'
    elif re.search(minus, s):
        sign = '-'
    if sign:
        if s[0] in ['+', '-']:
            dtm = ''
            dur = s[1:]
        else:
            parts = [x.strip() for x in re.split(r'[+-]\s', s)]
            dtm = parts[0]
            dur = parts[1] if len(parts) > 1 else ''
    else:
        dtm = s.strip()
        dur = ''
    if dtm in date_shortcuts:
        dt = date_shortcuts[dtm]()
    else:
        dt = parse(dtm)
    if dur:
        ok, du = parse_duration(dur)
    else:
        du = ZERO
    return dt + du if sign == '+' else dt - du


def _fmtdt(dt):
    # assumes dt is either a date or a datetime
    try:
        # this works if dt is a datetime
        # return dt.format("%YMMDDHHmm")
        return dt.strftime('%Y%m%d%H%M')
    except:
        # this works if dt is a date by providing 00 for HH and mm
        # return dt.format("%YMMDD0000")
        return dt.strftime('%Y%m%d0000')


def later(d1, d2):
    """
    True if d1 > d2
    """
    if not (isinstance(d1, date) and isinstance(d2, date)):
        # each must be a date or a datetime
        return 'only dates or datetimes can be compared'
    return _fmtdt(d1) > _fmtdt(d2)


def earlier(d1, d2):
    """
    True if d1 < d2
    """
    if not (isinstance(d1, date) and isinstance(d2, date)):
        # each must be a date or a datetime
        return 'only dates or datetimes can be compared'
    return _fmtdt(d1) < _fmtdt(d2)


def maybe_round(obj):
    """
    round up to the nearest UT_MIN minutes.
    """

    if isinstance(obj, Period):
        obj = obj.diff
    if not isinstance(obj, timedelta):
        return None
    if UT_MIN == 0:
        return obj
    seconds = int(obj.total_seconds()) % 60
    if seconds:
        # round up to the next minute
        obj = obj + timedelta(seconds=60 - seconds)

    res = (obj.total_seconds() // 60) % UT_MIN
    if res:
        obj += (UT_MIN - res) * ONEMIN
    return obj


def sort_dates_times(obj):
    if isinstance(obj, date):
        return obj.strftime('%Y-%m-%d 00:00')
    if isinstance(obj, dateTime):
        return obj.strftime('%Y-%m-%d %H:%M')
    else:
        return obj


def apply_dates_filter(items, grpby, filters):
    if grpby['report'] in ['u', 'v']:

        def rel_dt(item, filters):
            rdts = []
            ok = False
            used_times = deepcopy(item['u'])
            drop = []
            if 'b' in filters and 'e' in filters:   # both b and e
                for x in used_times:
                    if earlier(x[1], filters['b']) or later(
                        x[1], filters['e']
                    ):
                        drop.append(x)
            elif 'b' in filters:   # only b
                for x in used_times:
                    if earlier(x[1], filters['b']):
                        drop.append(x)
            elif 'e' in filters:   # only e
                for x in used_times:
                    if later(x[1], filters['e']):
                        drop.append(x)
            if drop:
                for x in drop:
                    used_times.remove(x)
            items = []
            dt2ut = {}
            for x in used_times:
                rdt = x[1] if isinstance(x[1], date) else x[1].date()
                dt2ut.setdefault(rdt, ZERO)
                dt2ut[rdt] += maybe_round(x[0])
            for rdt in dt2ut:
                tmp = deepcopy(item)
                tmp['rdt'] = rdt
                tmp['u'] = [rdt, dt2ut[rdt]]
                items.append(tmp)
            return items

    elif grpby['report'] == 's':

        def rel_dt(item, filters):
            ok = False
            items = []
            tmp = deepcopy(item)
            rdt = None
            if 'f' in item:
                rdt = (
                    item['f']
                    if isinstance(item['f'], date)
                    else item['f'].date()
                )
                # e_ok = 'e' not in filters or item['f'] <= filters['e']
                e_ok = 'e' not in filters or not later(item['f'], filters['e'])
                # b_ok = 'b' not in filters or item['f'] >= filters['b']
                b_ok = not (
                    'b' in filters and earlier(item['f'], filters['b'])
                )
            elif 's' in item:
                rdt = (
                    item['s']
                    if isinstance(item['s'], date)
                    else item['s'].date()
                )
                # e_ok = 'e' not in filters or rdt <= filters['e']
                e_ok = not ('e' in filters and later(item['s'], filters['e']))
                # b_ok = 'b' not in filters or rdt >= filters['b']
                b_ok = 'b' not in filters or not earlier(
                    item['s'], filters['b']
                )
            else:
                e_ok = b_ok = True
            if e_ok and b_ok:
                if grpby['dated']:
                    if rdt:
                        tmp['rdt'] = rdt
                        items.append(tmp)
                else:
                    # not dated, don't need rdt
                    items.append(tmp)
            return items

    elif grpby['report'] == 'c':

        def rel_dt(item, filters):
            ok = False
            items = []
            tmp = deepcopy(item)
            rdt = item['created']
            # e_ok = 'e' not in filters or rdt <= filters['e']
            e_ok = 'e' not in filters or not later(rdt, filters['e'])
            # b_ok = 'b' not in filters or rdt >= filters['b']
            b_ok = 'b' not in filters or not earlier(rdt, filters['b'])
            if e_ok and b_ok:
                if rdt:
                    tmp['rdt'] = rdt
                    items.append(tmp)
                else:
                    # not dated, don't need rdt
                    items.append(tmp)
            return items

    elif grpby['report'] == 'm':

        def rel_dt(item, filters):
            ok = False
            items = []
            tmp = deepcopy(item)
            rdt = item.get('modified', item['created'])
            # e_ok = 'e' not in filters or rdt <= filters['e']
            e_ok = 'e' not in filters or not later(rdt, filters['e'])
            # b_ok = 'b' not in filters or rdt >= filters['b']
            b_ok = 'b' not in filters or not earlier(rdt, filters['b'])
            if e_ok and b_ok:
                if rdt:
                    tmp['rdt'] = rdt
                    items.append(tmp)
                else:
                    # not dated, don't need rdt
                    items.append(tmp)
            return items

    ok_items = []
    for item in items:
        ok_items.extend(rel_dt(item, filters))
    return ok_items


class QDict(dict):
    """
    Constructed from rows of (path, values) tuples. The path will be split using 'split_char' to produce the nodes leading to 'values'. The last element in values is presumed to be the 'id' of the item that generated the row.
    """

    tab = 2

    def __init__(self, used_time={}, row=0, needs=[]):
        self.width = shutil.get_terminal_size()[0] - 4   # -2 for query indent
        self.row = row
        self.needs = needs
        self.row2id = {}
        self.output = []
        self.used_time = used_time

        self.fmt = ''
        if self.needs:
            fmts = {'y': '%y', 'm': '%-m', 'd': '%-d'}
            tmp = []
            if settings['yearfirst'] and not settings['dayfirst']:
                for x in ['y', 'm', 'd']:
                    if x in self.needs:
                        tmp.append(fmts[x])
            else:
                for x in ['d', 'm', 'y']:
                    if x in self.needs:
                        tmp.append(fmts[x])
            self.fmt = '/'.join(tmp)
            if settings['ampm']:
                self.fmt += ' %-I:%M%p'
            else:
                self.fmt += '%H:%M'

    def __missing__(self, key):
        self[key] = QDict()
        return self[key]

    def as_dict(self):
        return self

    def leaf_detail(self, detail, depth):
        dindent = QDict.tab * (depth + 1) * ' '
        if isinstance(detail, str):
            paragraphs = detail.split('\n')
            ret = []
            for para in paragraphs:
                ret.extend(
                    textwrap.fill(
                        para,
                        initial_indent=dindent,
                        subsequent_indent=dindent,
                        width=self.width - QDict.tab * (depth - 1),
                    ).split('\n')
                )
        elif isinstance(detail, datetime):
            ret = [dindent + format_datetime(detail, short=True)[1]]
        elif isinstance(detail, timedelta):
            ret = [dindent + format_hours_and_tenths(detail)]
        elif isinstance(detail, list) and detail:
            if isinstance(detail[0], str):
                ret = [dindent + ', '.join(detail)]
            elif isinstance(detail[0], list):
                # u, e.g., will be a list of duration, datetime tuples
                ret = []
                detail.sort(key=lambda x: sort_dates_times(x[1]))
                for d in detail:
                    try:
                        tmp = f'{format_hours_and_tenths(d[0])}: {format_datetime(d[1], short=True)[1]}'
                        ret.append(dindent + tmp)
                    except Exception as e:
                        # logger.error(f'error {e}, processing {d}')
                        ret.append(dindent + repr(d))
            else:
                ret = []
                try:
                    tmp = f'{format_datetime(detail[0], short=True)[1]}: {format_hours_and_tenths(detail[1])}'
                    ret.append(dindent + tmp)
                except Exception as e:
                    # logger.error(f'error {e}, processing {detail}')
                    ret.append(dindent + repr(detail))

        else:
            ret = [dindent + repr(detail)]
        return ret

    def add(self, keys, values=()):
        for j in range(len(keys)):
            key = keys[j]
            keys_left = keys[j + 1 :]
            if not keys_left:
                try:
                    self.setdefault(key, []).append(values)
                except Exception as e:
                    logger.warning(
                        f'error adding key: {key}, values: {values}\n self: {self}; e: {repr(e)}'
                    )
            if isinstance(self[key], dict):
                self = self[key]
            elif keys_left:
                self.setdefault(':'.join(keys[j:]), []).append(values)
                break

    def as_tree(self, t={}, depth=0, level=0, pre=[]):
        """return an indented tree"""
        for k in t.keys():
            del pre[depth:]
            pre.append(k)
            indent = QDict.tab * depth * ' '
            if self.used_time:
                self.output.append(
                    '%s%s %s'
                    % (
                        indent,
                        k,
                        format_hours_and_tenths(
                            self.used_time.get(tuple(pre), '')
                        ),
                    )
                )
            else:
                self.output.append('%s%s' % (indent, k))
            self.row += 1
            depth += 1
            if level and depth > level:
                depth -= 1
                continue

            if type(t[k]) == QDict:
                self.as_tree(t[k], depth, level, pre)
            else:
                for leaf in t[k]:
                    indent = QDict.tab * depth * ' '
                    l_indent = len(indent)
                    ut = format_hours_and_tenths(leaf[2])
                    # dt = leaf[3].strftime('%d %H:%M')
                    # dt = format_datetime(leaf[3], short=True)[1]
                    dt = leaf[3].strftime(self.fmt).rstrip('M').lower()
                    avail = self.width - l_indent - len(ut) - len(dt) - 4
                    if len(leaf[1]) > avail:
                        leaf[1] = (
                            leaf[1][: avail - 1] + ETM_CHAR['ELLIPSiS_CHAR']
                            if len(leaf[1]) >= avail
                            else leaf[1]
                        )
                    if self.used_time:
                        self.output.append(
                            f'{indent}{leaf[0]} {ut} {dt} {leaf[1]}'
                            # '%s%s %s: %s %s'
                            # % (indent, leaf[0], dt, leaf[1], ut)
                        )
                        num_leafs = 4
                    else:
                        self.output.append(
                            '%s%s %s' % (indent, leaf[0], leaf[1])
                        )
                        num_leafs = 2
                    self.row2id[self.row] = leaf[-1]
                    self.row += 1
                    # len(leaf) - 1 skips the last integer doc_id leaf
                    for i in range(num_leafs, len(leaf) - 1):
                        lines = self.leaf_detail(leaf[i], depth)
                        for line in lines:
                            self.output.append(line)
                            self.row += 1

            depth -= 1
        return '\n  '.join(self.output), self.row2id


def get_output_and_row2id(items, grpby, header='', needs=[]):
    used_time = {}
    ret = []
    report = grpby['report']
    sort_tups = [x for x in grpby.get('sort', [])]
    path_tups = [x for x in grpby.get('path', [])]
    dtls_tups = [x for x in grpby.get('dtls', [])]
    mode = header[0]
    for item in items:
        for x in ['i', 'c', 'l']:
            item.setdefault(x, '~')   # make ~ the default
        item.setdefault('modified', item['created'])
        if 'f' in item:
            item['itemtype'] = '-' if report in ['u', 'v'] else finished_char
        st = [
            eval(x, {'item': item, 're': re, 'format_week': format_week})
            for x in sort_tups
            if x
        ]
        pt = []
        for y in path_tups:
            if not y:
                continue
            if isinstance(y, list):
                pt.append(
                    ' '.join(
                        [
                            eval(
                                x,
                                {
                                    'item': item,
                                    're': re,
                                    'format_week': format_week,
                                },
                            )
                            for x in y
                            if x
                        ]
                    )
                )
            else:
                pt.append(
                    eval(
                        y, {'item': item, 're': re, 'format_week': format_week}
                    )
                )

        dt = []
        for x in dtls_tups:
            if not x:
                continue
            try:
                dt.append(eval(x, {'item': item, 're': re}))
            except Exception as e:
                logger.error(f'error: {e}, evaluating {x}')
        if grpby['report'] in ['u', 'v']:
            dt[2] = ut = maybe_round(dt[2])
            for i in range(len(pt)):
                key = tuple(pt[: i + 1])
                used_time.setdefault(key, ZERO)
                used_time[key] += ut
        ret.append((st, pt, dt))
    ret.sort(key=lambda x: x[0])

    # drop the sort key
    ret = [x[1:] for x in ret]
    # logger.debug(ret)

    if mode == 'v':
        # CVS stuff
        now = datetime.now().astimezone().strftime('%y%m%dT%H%M%S')
        # txtfile = os.path.join(csvdir, f'{now}.txt')
        csvfile = os.path.join(csvdir, f'{now}.csv')
        count = 0
        txt_output = [
            f'{count}) {header}',
        ]
        # num_fields = len(ret[0][1]) + 1
        # header_row = ['' for x in range(num_fields)]
        header_row = ['#', 'type', 'summary', 'hours', 'datetime']
        for x in dtls_tups[4:-1]:
            header_row.append(x)
        header_row.append('doc_id')
        header_row.append(header)
        # logger.debug(f'num_fields: {num_fields}, header_row: {header_row}')
        csv_output = [
            header_row,
        ]
        last_path = []
        for row in ret:
            count += 1
            path, body = row
            # logger.debug(f'path: {path}, body: {body}')
            if path != last_path:
                last_path = path
                # txt_output.append(f"{count}) {' '.join(path)}")
                append = ' / '.join(path)
            else:
                append = ''
            tmp = [
                f'{count}',
            ]
            for part in body:
                # logger.debug(f'processing {part} {type(part)}')
                if isinstance(part, str):
                    tmp.append(part)
                elif isinstance(part, int):
                    tmp.append(str(part))
                elif isinstance(part, timedelta):
                    tmp.append(format_hours_and_tenths(part).rstrip('h'))
                elif isinstance(part, datetime):
                    tmp.append(part.strftime('%Y-%m-%d %H:%M'))
                else:
                    tmp.append(part.__repr__)
            tmp.append(append)
            csv_output.append(tmp)
        with open(csvfile, 'w+') as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerows(csv_output)
        # with open(txtfile, 'w') as txt_file:
        #     txt_file.write('\n'.join(txt_output))
        return (
            f'csv output saved to {csvfile}',
            {},
        )
    else:
        header = textwrap.fill(
            f'query: {header}',
            width=shutil.get_terminal_size()[0] - 2,
            initial_indent='',
            subsequent_indent='   ',
        )
        # create recursive dict from data
        row = 1 if header else 0
        # need to pass 'needs' to QDict
        index = QDict(used_time, row, needs)
        for path, value in ret:
            index.add(path, value)

        output, row2id = index.as_tree(index)
        return f'{header}\n  {output}', row2id


def get_grpby_and_filters(s, options=None):
    if not options:
        options = {}
    grpby = {}
    filters = {}
    op_str = s.split('#')[0]
    parts = minus_regex.split(op_str)
    head = parts.pop(0)
    report = head[0]
    groupbystr = head[1:].strip()
    if not (report and report in ['s', 'u', 'm', 'c', 'v']):
        return {}, {}
    grpby['report'] = report
    filters['dates'] = False
    grpby['dated'] = False
    filters['query'] = (
        'exists u' if report in ['u', 'v'] else 'exists itemtype'
    )
    groupbylst = []
    if groupbystr:
        groupbylst = [x.strip() for x in groupbystr.split(';')]
        for comp in groupbylst:
            comp_parts = comp.split(' ')
            for part in comp_parts:
                if groupdate_regex.match(part):
                    grpby['dated'] = True
                    filters['dates'] = True
                elif not (
                    part[0] == 'i'
                    or part
                    in ['c', 'l', 'itemtype', 'created', 'modified', 'summary']
                ):
                    print(f'Ignoring invalid groupby part: {part}')
                    groupbylst.remove(comp)
        grpby['args'] = groupbylst
    grpby['path'] = []
    grpby['sort'] = []
    # include = {'W', 'Y', 'M', 'D'}
    # needs = ['%y', '%-m', '%-d']
    needs = ['y', 'm', 'd']
    if groupbylst:
        if grpby['dated'] or grpby['report'] in ['u', 'v', 'm', 'c']:
            grpby['sort'].append(f"item['rdt'].strftime('%Y%m%d')")
        for group in groupbylst:
            if groupdate_regex.search(group):
                gparts = group.split(' ')
                this_sort = []
                this_path = []
                for part in gparts:
                    if 'W' in part:
                        this_sort.append("item['rdt'].strftime('%W')")
                        this_path.append(
                            f"format_week(item['rdt'], {p2d[part]})"
                        )
                    if 'Y' in part:
                        needs.remove('y')
                        this_sort.append("item['rdt'].strftime('%Y')")
                        this_path.append(
                            f"item['rdt'].strftime('{p2d[part]}')"
                        )
                    if 'M' in part:
                        needs.remove('m')
                        this_sort.append("item['rdt'].strftime('%m')")
                        this_path.append(
                            f"item['rdt'].strftime('{p2d[part]}')"
                        )
                    if 'D' in part:
                        needs.remove('d')
                        this_sort.append("item['rdt'].strftime('%d')")
                        this_path.append(
                            f"item['rdt'].strftime('{p2d[part]}')"
                        )
                    if 'd' in part:
                        this_path.append(
                            f"item['rdt'].strftime('{p2d[part]}')"
                        )
                grpby['sort'].extend(this_sort)
                grpby['path'].append(this_path)

            elif '[' in group and group[0] == 'i':
                if ':' in group:
                    grpby['path'].append(
                        f"'/'.join(re.split('/', item['{group[0]}']){group[1:]}) or '~'"
                    )
                    grpby['sort'].append(
                        f"'/'.join(re.split('/', item['{group[0]}']){group[1:]}) or '~'"
                    )
                # else:
                #     # logger.warning(f'non slice use of i: {group}')
            else:
                grpby['path'].append("item['%s']" % group.strip())
                grpby['sort'].append(f"item['{group.strip()}']")

    also = []
    for part in parts:
        key = part[0]
        if key == 'a':
            value = [x.strip() for x in part[1:].split(',')]
            also.extend(value)
        elif key in ['b', 'e']:
            # dt = parse(part[1:], strict=False, tz='local')
            dt = parse_reldt(part[1:])
            filters[key] = dt
        elif key == 'q':
            value = part[1:].strip()
            filters['query'] += f' and {value}'
        elif key == 't':
            value = [x.strip() for x in part[1:].split(',')]
    details = ["item['itemtype']", "item['summary']"]
    if report in ['u', 'v']:
        details.append("item['u'][1]")
        details.append("item['u'][0]")
    else:
        details.append('')
    if also:
        details.extend([f"item.get('{x}', '~')" for x in also])
    details.append('item.doc_id')
    grpby['dtls'] = details
    return grpby, filters, needs


def show_query_results(text, grpby, items, needs):
    # called by dataview to set query
    rows = []
    item_count = f'[{len(items)} records]'
    if not (items and isinstance(items, list)):
        return f'query: {text}\n   none matching', {}
    # header = textwrap.fill(
    #     f'query: {text} {item_count}',
    #     width=width,
    #     initial_indent='# ',
    #     subsequent_indent='# ',
    # )
    header = f'{text} {item_count}'
    mode = text[0]
    # need to pass 'needs' to get_output_and_row2id
    output, row2id = get_output_and_row2id(items, grpby, header, needs)
    return output, row2id


def main(etmdir, args):

    # from etm.view import Query
    # logger = setup_logging(1, etmdir)
    # from etm.options import Settings

    # settings = Settings(etmdir)
    query = ETMQuery()

    session = PromptSession()

    again = True
    while again:

        print("Enter 'quit', 'stop', 'exit' or '' to exit loop")
        text = session.prompt('query: ', lexer=query.lexer)
        if not text or text in ['quit', 'stop', 'exit']:
            again = False
            continue
        if text == 'd':   # with descriptions
            text = 'u i[0]; MMM YYYY; i[1:]; ddd D -b 3/1 -e 4/1 -a d'
        elif text == 'p':   # without descriptions
            text = 'u i[0]; MMM YYYY; i[1:]; ddd D -b 3/1 -e 4/1'
        elif text == 'w':   # without descriptions and days
            text = 'u i[0]; MMM YYYY; i[1:] -b 1/1 -e 3/1'
        elif text == 'a':   # client a only
            text = 'u i[0]; MMM YYYYY; i[1:] -q matches i client\\sa -b 1/1 -e 3/1 -a d'
        elif len(text.strip()) == 1:
            print('missing report arguments')
            continue
        print(f'query: {text}')
        if (
            len(text) > 1
            and text[1] == ' '
            and text[0] in ['s', 'u', 'm', 'c']
        ):
            grpby, filters = get_grpby_and_filters(text)
            if not grpby:
                continue
            print(f'grpby: {grpby}\n\nfilters: {filters}')
            ok, items = query.do_query(filters.get('query'))
            if ok:
                items = apply_dates_filter(items, grpby, filters)
                output, row2id = get_output_and_row2id(items, grpby)
                print(output)
            else:
                print(items)
        else:
            ok, items = query.do_query(text)
            if ok:
                for item in items:
                    print(
                        f"   {item['itemtype']} {item.get('summary', 'none')} {item.doc_id}"
                    )
            else:
                print(items)


if __name__ == '__main__':
    print('This module should only be imported')
    sys.exit()
    # main(*sys.argv[1:])



#! ./env/bin/python
from datetime import datetime, date, timedelta
from prompt_toolkit import prompt
from prompt_toolkit.validation import Validator, ValidationError
import logging
import logging.config

import sys
import os


class ConfirmationValidator(Validator):
    def validate(self, document):
        if document.text.lower() not in ('y', 'n'):
            raise ValidationError(
                message='Please enter either "y" or "n"',
                cursor_position=len(document.text),
            )


def ask_for_confirmation(prompt_message):
    while True:
        response = prompt(
            f'{prompt_message} [y/n]: ', validator=ConfirmationValidator()
        )
        if response.lower() == 'y':
            return True
        elif response.lower() == 'n':
            return False


def print_usage():

    print(
        """\
    Usage: etm [options]
    Options:
      -h, --help   Show this help message and exit
      [n] [path]   Set logging level 'n' where n = 1, 2, 3
                       or, if omitted, use logging level 2
                   Use 'path' as the etm home directory
                       or, if omitted, use the environmental
                       variable ETMHOME if set and the current
                       working directory otherwise."""
    )


ETMHOME = ''


def setup_logging(level, etmdir, file=None):
    """
    Setup logging configuration. Override root:level in
    logging.yaml with default_level.
    """

    if not os.path.isdir(etmdir):
        return

    log_levels = {
        1: logging.DEBUG,
        2: logging.INFO,
        3: logging.WARN,
        4: logging.ERROR,
        5: logging.CRITICAL,
    }

    level = int(level)
    loglevel = log_levels.get(level, log_levels[3])

    # if we get here, we have an existing etmdir
    logfile = os.path.normpath(
        os.path.abspath(os.path.join(etmdir, 'etm.log'))
    )

    config = {
        'disable_existing_loggers': False,
        'formatters': {
            'simple': {
                'format': '--- %(asctime)s - %(levelname)s - %(module)s.%(funcName)s\n    %(message)s'
            }
        },
        'handlers': {
            'file': {
                'backupCount': 7,
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'encoding': 'utf8',
                'filename': logfile,
                'formatter': 'simple',
                'level': loglevel,
                'when': 'midnight',
                'interval': 1,
            }
        },
        'loggers': {
            'etmmv': {
                'handlers': ['file'],
                'level': loglevel,
                'propagate': False,
            }
        },
        'Redirectoot': {'handlers': ['file'], 'level': loglevel},
        'version': 1,
    }
    logging.config.dictConfig(config)
    # logger = logging.getLogger('asyncio').setLevel(logging.WARNING)
    logger = logging.getLogger('etmmv')

    logger.critical('\n######## Initializing logging #########')
    if logfile:
        logger.critical(
            f'logging for file: {file}\n    logging at level: {loglevel}\n    logging to file: {logfile}'
        )
    else:
        logger.critical(
            f'logging at level: {loglevel}\n    logging to file: {logfile}'
        )
    return logger


def main():

    global ETMHOME

    from etm.common import TimeIt

    if '-h' in sys.argv or '--help' in sys.argv:
        print_usage()
        sys.exit()

    if len(sys.argv) > 1 and sys.argv[1] in ['-h', '--h', '-help', '--help']:
        print(
            """
Usage:
    etm <

              """
        )
        sys.exit()

    MIN_PYTHON = (3, 7, 3)
    if sys.version_info < MIN_PYTHON:
        mv = '.'.join([str(x) for x in MIN_PYTHON])
        sys.exit(f'Python {mv} or later is required.\n')
    import os

    IS_VENV = os.getenv('VIRTUAL_ENV') is not None

    import etm.__version__ as version

    etm_version = version.version
    etmhome = os.environ.get('ETMHOME')
    etmdir = etmhome if etmhome else os.getcwd()

    import etm.data as data
    from etm.data import Period
    import etm.view as view

    from etm.view import ETMQuery
    import etm.model as model
    import etm.common as common

    import etm.report as report

    report.ETMQuery = ETMQuery

    loglevel = 2   # info
    log_levels = [str(x) for x in range(1, 6)]
    if len(sys.argv) > 1 and sys.argv[1] in log_levels:
        loglevel = int(sys.argv.pop(1))
    if len(sys.argv) > 1 and sys.argv[1] in ['model', 'view', 'data', 'rep']:
        if sys.argv[1] == 'model':
            logger.info(
                f'calling model doctest with etmdir: {etmdir}, argv: {sys.argv}'
            )
            import doctest

            doctest.testmod(model)
        elif sys.argv[1] == 'view':
            logger.info(
                f'calling view doctest with etmdir: {etmdir}, argv: {sys.argv}'
            )
            import doctest

            doctest.testmod(view)
        elif sys.argv[1] == 'data':
            logger.info(
                f'calling data doctest with etmdir: {etmdir}, argv: {sys.argv}'
            )
            import doctest

            doctest.testmod(data)
        # elif sys.argv[1] == 'rep':
        #     logger.info(f"calling report.main with etmdir: {etmdir}, argv: {sys.argv}")
        #     report.main(etmdir, sys.argv)
        sys.exit()
        # else:
        #     logger.info(f"calling model.main with etmdir: {etmdir}, argv: {sys.argv}")
        #     model.main(etmdir, sys.argv)
    if len(sys.argv) > 1:
        # use the directory being provided
        etmdir = sys.argv.pop(1)
        etmdir = os.getcwd() if etmdir == '.' else os.path.normpath(etmdir)

    created_etmdir = False
    # if not (os.path.exists(etmdir) and os.path.isdir(etmdir)):
    if not os.path.isdir(etmdir):
        print(
            f"""
The provided directory to use for etm
    {etmdir}
does not exist and will need to be created.
              """
        )
        if ask_for_confirmation('Do you want to continue?'):
            print('Continuing...')
            os.makedirs(etmdir)
            created_etmdir = True
            print(f'created {etmdir}')
        else:
            print('Exiting...')
            sys.exit()

    logdir = os.path.normpath(os.path.join(etmdir, 'logs'))
    csvdir = os.path.normpath(os.path.join(etmdir, 'csv'))
    backdir = os.path.normpath(os.path.join(etmdir, 'backups'))
    cfgfile = os.path.normpath(os.path.join(etmdir, 'cfg.yaml'))
    dbfile = os.path.normpath(os.path.join(etmdir, 'etm.json'))
    missing = []
    for p in [logdir, backdir, csvdir, cfgfile, dbfile]:
        if not os.path.exists(p):
            missing.append(p)
    missing = '\n    '.join(missing) if missing else ''

    condition = not created_etmdir and missing

    if condition:
        print(
            f"""\
The etm directory
     {etmdir}
is missing
    {missing}
which will need to be created.
"""
        )
        if ask_for_confirmation('Do you want to continue?'):
            print('Continuing...')
        else:
            print('Exiting...')
            sys.exit()

    olddb = os.path.normpath(os.path.join(etmdir, 'db.json'))
    needs_update = False
    if os.path.exists(olddb) and not os.path.exists(dbfile):
        import shutil

        shutil.copy2(olddb, dbfile)
        needs_update = True

    if not os.path.isdir(logdir):
        os.makedirs(logdir)

    if not os.path.isdir(backdir):
        os.makedirs(backdir)

    if not os.path.isdir(csvdir):
        os.makedirs(csvdir)

    import etm.options as options

    setup_logging(loglevel, logdir)
    # logger = logging.getLogger('asyncio').setLevel(logging.WARNING)
    logger = logging.getLogger('etmmv')

    common.logger = logger
    options.logger = logger

    Settings = options.Settings(etmdir)

    settings = Settings.settings
    beginbusy = settings['beginbusy']
    type_colors = settings['type_colors']
    # logger.debug(f"__main__ type_colors: {type_colors}")
    window_colors = settings['window_colors']
    # logger.debug(f"__main__ window_colors: {window_colors}")

    logger.info(f'running in a virtual environment: {IS_VENV}')

    secret = settings.get('secret')
    queries = settings.get('queries')
    UT_MIN = settings.get('usedtime_minutes', 1)
    usedtime_hours = settings.get('usedtime_hours', 6)
    refresh_interval = settings.get('refresh_interval', 60)
    today = date.today()
    now = datetime.now().astimezone()
    sunday = today + timedelta(days=6 - today.weekday())   # sunday
    # We want 2 char 'en' weekday abbreviations regardless of the actual locale
    WA = {
        i: (sunday + i * timedelta(days=1)).strftime('%a')[:2]
        for i in range(1, 8)
    }
    midnight = datetime.now().replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    beginbusy = settings.get('beginbusy', 7)
    ampm = settings.get('ampm', True)
    show_minutes = settings.get('show_minutes', False)
    hour = timedelta(hours=1)
    fmt = '%I%p' if ampm else '%H'
    busyhours = [
        (midnight + i * hour).strftime('%-I%p').rstrip('M').lower()
        for i in range(0, 24, 6)
    ]
    HB = ''.join([f'{h : <8}' for h in busyhours]).rstrip()
    #
    # import etm.ical as ical
    # ical.logger = logger
    data.secret = secret
    data.logger = logger
    data.settings = settings
    from etm.data import Mask

    ETMDB = data.initialize_tinydb(dbfile)
    logger.info(f'initialized TinyDB using {dbfile}')
    DBITEM = ETMDB.table('items', cache_size=30)
    DBARCH = ETMDB.table('archive', cache_size=30)
    logger.debug(f'ETMDB: {ETMDB}, number of items: {len(ETMDB)}')

    from etm.make_examples import make_examples
    from etm.model import about
    from etm.model import parse
    from etm.model import wrap
    from etm.model import import_file
    from etm.model import import_examples
    from etm.model import write_back
    from etm.model import duration_in_words

    # from etm.model import RDict
    from etm.model import item_details

    # from etm.model import FINISHED_CHAR
    # from etm.model import UPDATE_CHAR
    # from etm.model import PIN_CHAR
    # from etm.model import INBASKET_CHAR

    # ical.parse = parse
    model.loglevel = loglevel
    model.etm_version = etm_version
    model.secret = secret
    model.data = data
    model.Period = Period
    # model.ical = ical
    model.Mask = Mask
    model.WA = WA
    model.HB = HB
    # model.VSEP = VSEP
    # model.HSEP = HSEP
    # model.BUSY = BUSY
    # model.CONF = CONF
    # model.ADAY = ADAY
    # model.USED = USED
    # model.FREE = FREE
    # model.REPS = REPS
    model.ETMDB = ETMDB
    model.DBITEM = DBITEM
    model.DBARCH = DBARCH
    model.UT_MIN = UT_MIN
    model.usedtime_hours = usedtime_hours
    model.refresh_interval = refresh_interval
    model.beginbusy = beginbusy
    model.settings = settings
    common.settings = settings
    model.logger = logger
    model.make_examples = make_examples
    model.needs_update = needs_update
    model.timers_file = os.path.join(etmdir, 'timers.pkl')
    userhome = os.path.expanduser('~')
    etmhome = (
        os.path.join('~', os.path.relpath(etmdir, userhome))
        if etmdir.startswith(userhome)
        else etmdir
    )
    logger.debug(f'etmhome: {etmhome}')
    # common.etmhome = etmhome
    model.etmhome = etmhome
    ETMHOME = etmhome
    # we put settings into the model namespace so model.Dataview will have it
    dataview = model.DataView(etmdir)
    datetime_calculator = model.datetime_calculator
    item = model.Item()
    format_time = model.format_time
    format_datetime = model.format_datetime
    format_statustime = model.format_statustime
    format_duration = model.format_duration
    format_hours_and_tenths = model.format_hours_and_tenths
    # since dataview calls schedule it will also have settings
    completions = dataview.completions
    expansions = settings['expansions']
    if expansions:
        for x in expansions:
            completions.append(f'@x {x}')
    style = dataview.settings['style']
    parse_datetime = model.parse_datetime
    parse_duration = model.parse_duration

    view.loglevel = loglevel
    view.TimeIt = TimeIt
    view.wrap = wrap
    view.parse = parse
    view.WA = WA
    view.beginbusy = beginbusy
    # view.FINISHED_CHAR = FINISHED_CHAR
    # view.UPDATE_CHAR = UPDATE_CHAR
    # view.PIN_CHAR = PIN_CHAR
    # view.INBASKET_CHAR = INBASKET_CHAR
    view.settings = settings
    view.type_colors = type_colors
    view.cfgfile = cfgfile
    view.model = model
    view.duration_in_words = duration_in_words
    view.write_back = write_back
    # view.RDict = RDict
    # view.TDBLexer = model.TBDLexer
    view.item = item
    view.item_details = item_details
    view.logger = logger
    view.import_file = import_file
    view.import_examples = import_examples
    view.etmdir = etmdir
    view.text_pattern = os.path.join(etmdir, '*.text')
    view.etmhome = etmhome
    view.datetime_calculator = datetime_calculator
    view.about = about
    view.format_time = format_time
    view.format_datetime = format_datetime
    view.format_statustime = format_statustime
    view.format_duration = format_duration
    view.parse_datetime = parse_datetime
    view.ETMDB = ETMDB
    view.DBITEM = DBITEM
    view.DBARCH = DBARCH
    view.etm_version = etm_version

    view.dataview = dataview
    view.completions = completions
    view.expansions = expansions
    view.terminal_style = style
    view.make_examples = make_examples

    view.report = report
    show_query_results = report.show_query_results
    view.show_query_results = show_query_results
    model.show_query_results = show_query_results
    # get_usedtime_cvs = report.get_usedtime_cvs
    # view.get_usedtime_cvs = get_usedtime_cvs
    # model.get_usedtime_cvs = get_usedtime_cvs

    report.ETMDB = ETMDB
    report.DBITEM = DBITEM
    report.DBARCH = DBARCH
    report.ETMQuery = ETMQuery
    report.settings = settings
    report.format_time = format_time
    report.parse_duration = parse_duration
    report.parse_datetime = parse_datetime
    report.format_datetime = format_datetime
    report.format_duration = format_duration
    report.format_hours_and_tenths = format_hours_and_tenths
    report.logger = logger
    report.UT_MIN = UT_MIN
    report.csvdir = csvdir

    logger.info(f'setting terminal_style: {style}')

    if len(sys.argv) > 1:
        if sys.argv[1] == 'model':
            logger.info(
                f'calling model doctest with etmdir: {etmdir}, argv: {sys.argv}'
            )
            import doctest

            doctest.testmod(model)
        elif sys.argv[1] == 'view':
            logger.info(
                f'calling view doctest with etmdir: {etmdir}, argv: {sys.argv}'
            )
            import doctest

            doctest.testmod(view)
        elif sys.argv[1] == 'data':
            logger.info(
                f'calling data doctest with etmdir: {etmdir}, argv: {sys.argv}'
            )
            import doctest

            doctest.testmod(data)
        # elif sys.argv[1] == 'rep':
        #     logger.info(f"calling report.main with etmdir: {etmdir}, argv: {sys.argv}")
        #     report.main(etmdir, sys.argv)
        else:
            logger.info(
                f'calling model.main with etmdir: {etmdir}, argv: {sys.argv}'
            )
            model.main(etmdir, sys.argv)

    else:
        logger.info(f'system info {model.about()[1]}')
        logger.info(f'calling view.main with etmdir: {etmdir}')

        # stderr to /dev/null
        fd = os.open('/dev/null', os.O_WRONLY)
        os.dup2(fd, 2)

        from etm.view import main
        import asyncio

        asyncio.run(main(etmdir), debug=True)


def inbasket():
    import sys

    typechar = '!'   # inbasket
    option = '@t etm+'

    help = f"""\
usage: etm+ 'text'          use text
   or: etm+                 get text from stdin
   or: etm+ [?|help]    print this usage information

With the environmental variable ETMHOME set to your etm
root directory, text either piped to this script or
provided as arguments will be appended to 'inbasket.text'
in the ETMHOME directory. When this file exists, etm will
display an ⓘ character at the right end of status bar
alerting you that this file is available for import by
pressing F5.

If the text provided to this script does not begin with an
etm typechar in  -, *, % or !, then the default typechar
'{typechar}' will be used. If the provided string does not
contain any @-key options, then '{option}' will be
appended.

If the inbox typechar '!' is used then after importing,
the reminder will appear as an 'inbox' item requiring your
attention in the list for the current day in agenda view.
This may be especially useful in composing quick notes with
the assurance that you will be reminded to sort them out
later.

If '{T}' is used anywhere in the input, it will be replaced
with a timestamp corresponding to the moment this script was
invoked.
"""

    etmhome = os.environ.get('ETMHOME')
    if not etmhome:
        print("The environmental variable 'ETMHOME' is missing but required.")
        sys.exit()
    elif not os.path.isdir(etmhome):
        print(
            f"The environmental variable 'ETMHOME={etmhome}' is not a valid directory."
        )
        sys.exit()

    inbasket = os.path.join(etmhome, 'inbasket.text')

    # use stdin if it's full
    if not sys.stdin.isatty():
        input = sys.stdin.read()
    # otherwise, get the input from
    elif len(sys.argv) > 1:
        if len(sys.argv) == 2:
            input = ' '.join(sys.argv[1:])
        else:
            print('The provided input should be wrapped in single quotes')
            sys.exit()
    else:
        print(help)
        sys.exit()

    input = input.strip()
    if input in ['help', '?']:
        print(help)
        sys.exit()

    # if input does not begins with an itemtype character, prepend typechar .
    input = input if input[0] in '!*-%' else f'{typechar} {input}'
    input = input if '@' in input else f'{input} {option}'
    if '{T}' in input:
        # expand the timestamp
        hsh = {
            'T': datetime.now().astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')
        }
        input = input.format(**hsh)

    with open(inbasket, 'a') as fo:
        fo.write(f'{input}\n')
    print(f"appended:\n   '{input}'\nto {inbasket}")


if __name__ == '__main__':
    # main()
    asyncio.get_event_loop().run_until_complete(main())



#!/usr/bin/env python3

import os
import json
from tinydb import JSONStorage, TinyDB, Storage
from tinydb.table import Table, Document
from tinydb.middlewares import CachingMiddleware
from contextlib import contextmanager
from typing import Dict, Any, Optional, Callable, Mapping

from tinydb import __version__ as tinydb_version
from tinydb_serialization import Serializer
from tinydb_serialization import SerializationMiddleware
import base64  # for do_mask

from datetime import datetime, date, timedelta

import dateutil
import dateutil.rrule
from zoneinfo import ZoneInfo

from dateutil.rrule import *
import re
from etm.common import (
    WKDAYS_DECODE,
    WKDAYS_ENCODE,
    AWARE_FMT,
    NAIVE_FMT,
    DATE_FMT,
)
from etm.common import Period, is_aware, encode_datetime, decode_datetime

##########################
### begin TinyDB setup ###
##########################


class DateTimeSerializer(Serializer):
    """
    This class handles both aware and naive datetime objects.

    Encoding: If obj.tzinfo is 'None', it is interpreted as naive, serialized without conversion and an 'N' is appended. Otherwise it is interpreted as aware, converted to UTC and an 'A' is appended.
    Decoding: If the serialization ends with 'A', the datetime object is treated as 'UTC' and converted to localtime. Otherwise, the object is treated as 'Factory' and no conversion is performed.

    This serialization discards both seconds and microseconds but preserves hours and minutes.

    """

    OBJ_CLASS = datetime

    def encode(self, obj):
        """
        Serialize naive objects (Z == '') without conversion but with 'N' for 'Naive' appended. Convert aware datetime objects to UTC and then serialize them with 'A' for 'Aware' appended.
        >>> dts = DateTimeSerializer()
        >>> dts.encode(datetime(2018,7,25,10,27).naive())
        '20180725T1027N'
        >>> dts.encode(datetime(2018,7,25,10,27, tz='US/Eastern'))
        '20180725T1427A'
        """
        return encode_datetime(obj)
        if is_aware(obj):
            return obj.astimezone(ZoneInfo('UTC')).strftime(AWARE_FMT)
        else:
            return obj.strftime(NAIVE_FMT)

    def decode(self, s):
        """
        Return the serialization as a datetime object. If the serializaton ends with 'A',  first interpret as UTC aware and then return the corresponding aware datetime object in the local timezone. If the serialization ends with 'N', returning as an aware datetime object in the local timezone.
        >>> dts = DateTimeSerializer()
        >>> dts.decode('20180725T1027N')
        DateTime(2018, 7, 25, 10, 27, 0, tzinfo=Timezone('America/New_York'))
        >>> dts.decode('20180725T1427A')
        DateTime(2018, 7, 25, 10, 27, 0, tzinfo=Timezone('America/New_York'))
        """
        if s[-1] == 'A':
            return (
                datetime.strptime(s, AWARE_FMT)
                .replace(tzinfo=ZoneInfo('UTC'))
                .astimezone()
            )
        else:
            return datetime.strptime(s, NAIVE_FMT).astimezone(None)


class DateSerializer(Serializer):
    """
    This class handles datetime date objects. Encode as date string and decode as a midnight datetime without conversion in the local timezone.
    >>> ds = DateSerializer()
    >>> ds.encode(datetime(2018, 7, 25).date())
    '20180725'
    >>> ds.decode('20180725')
    Date(2018, 7, 25)
    """

    OBJ_CLASS = date

    def encode(self, obj):
        """
        Serialize the naive date object without conversion.
        """
        return obj.strftime(DATE_FMT)

    def decode(self, s):
        """
        Return the serialization as a date object.
        """
        return datetime.strptime(s, DATE_FMT).date()


class PeriodSerializer(Serializer):
    """
    This class handles both aware and 'factory' datetime objects.

    Encoding: If obj.tzinfo.abbrev is '-00' (tz=Factory), it is interpreted as naive, serialized without conversion and an 'N' is appended. Otherwise it is interpreted as aware, converted to UTC and an 'A' is appended.
    Decoding: If the serialization ends with 'A', the datetime object is treated as 'UTC' and converted to localtime. Otherwise, the object is treated as 'Factory' and no conversion is performed.

    This serialization discards both seconds and microseconds but preserves hours and minutes.

    """

    OBJ_CLASS = Period

    def encode(self, obj):
        """
        Serialize naive objects (Z == '') without conversion but with 'N' for 'Naive' appended.
        Convert aware datetime objects to UTC and then serialize them with 'A' for 'Aware' appended.
        >>> ps = PeriodSerializer()
        >>> ps.encode(Period(datetime(2018,7,30,10,45).naive(), datetime(2018,7,25,10,27).naive()))
        '20180730T1045N -> 20180725T1027N'
        >>> ps.encode(Period(datetime(2018,7,30,10,45).astimezone(pytz.tzinfo('US/Eastern')), datetime(2018,7,25,10,27)astimezone(pytz('US/Pacific')))
        '20180730T1445A -> 20180725T1727A'
        """
        start_fmt = encode_datetime(obj.start)
        end_fmt = encode_datetime(obj.end)
        return f'{start_fmt} -> {end_fmt}'

    def decode(self, s):
        """
        Return the serialization as a period object. If the serializaton ends with 'A',  first converting to localtime and returning an aware datetime object in the local timezone. If the serialization ends with 'N', returning without conversion as an aware datetime object in the local timezone.
        >>> ps = PeriodSerializer()
        >>> ps.decode('20180730T1045N -> 20180725T1027N')
        <Period [2018-07-30T10:45:00-04:00 -> 2018-07-25T10:27:00-04:00]>
        >>> ps.decode('20180730T1445A -> 20180725T1727A')
        <Period [2018-07-30T10:45:00-04:00 -> 2018-07-25T13:27:00-04:00]>
        """

        start, end = [x.strip() for x in s.split('->')]
        start_enc = decode_datetime(start)
        end_enc = decode_datetime(end)
        return Period(start_enc, end_enc)

    def __repr__(self):
        return f'{self.obj}'


class DurationSerializer(Serializer):
    """
    This class handles timedelta (timedelta) objects.
    >>> dus = DurationSerializer()
    >>> dus.encode(timedelta(days=3, hours=5, minutes=15))
    '3d5h15m'
    >>> dus.decode('3d5h15m')
    Duration(days=3, hours=5, minutes=15)
    """

    OBJ_CLASS = timedelta

    def encode(self, obj):
        """
        Serialize the timedelta object as days.seconds.
        """
        return format_duration(obj)

    def decode(self, s):
        """
        Return the serialization as a timedelta object.
        """
        return parse_duration(s)[1]


class WeekdaySerializer(Serializer):
    """
    This class handles dateutil.rrule.weeday objects. Note in the following examples that
    unquoted weekdays, eg. MO(-3), are dateutil.rrule.weekday objects.
    >>> wds = WeekdaySerializer()
    >>> wds.encode(MO(-3))
    '-3MO'
    >>> wds.encode(SA(+3))
    '3SA'
    >>> wds.encode(WE)
    'WE'
    >>> wds.decode('-3MO')
    MO(-3)
    >>> wds.decode('3SA')
    SA(+3)
    >>> wds.decode('WE')
    WE
    """

    OBJ_CLASS = dateutil.rrule.weekday

    def encode(self, obj):
        """
        Serialize the weekday object.
        """
        s = WKDAYS_ENCODE[obj.__repr__()]
        if s.startswith('+'):
            # drop the leading + sign
            s = s[1:]
        return s

    def decode(self, s):
        """
        Return the serialization as a weekday object.
        """
        return eval('dateutil.rrule.{}'.format(WKDAYS_DECODE[s]))


########################################
###### Begin Mask ######################
########################################


def encode(key, clear):
    enc = []
    for i in range(len(clear)):
        key_c = key[i % len(key)]
        enc_c = chr((ord(clear[i]) + ord(key_c)) % 256)
        enc.append(enc_c)
    return base64.urlsafe_b64encode(''.join(enc).encode()).decode()


def decode(key, enc):
    dec = []
    enc = base64.urlsafe_b64decode(enc).decode()
    for i in range(len(enc)):
        key_c = key[i % len(key)]
        dec_c = chr((256 + ord(enc[i]) - ord(key_c)) % 256)
        dec.append(dec_c)
    return ''.join(dec)


# NOTE: The real secret is set in cfg.yaml
secret = 'whatever'


class Mask:
    """
    Provide an encoded value with an "isinstance" test for serializaton
    >>> mask = Mask('my dirty secret')
    >>> isinstance(mask, Mask)
    True
    """

    def __init__(self, message=''):
        self.encoded = encode(secret, message)

    def __repr__(self):
        return decode(secret, self.encoded)


class MaskSerializer(Serializer):
    """ """

    OBJ_CLASS = Mask

    def encode(self, obj):
        """
        Encode the string using base64
        """
        return obj.encoded

    def decode(self, s):
        """
        Decode the base
        """
        return Mask(decode(secret, s))


########################################
###### End Mask ########################
########################################


def initialize_tinydb(dbfile):
    """ """
    serialization = SerializationMiddleware()
    serialization.register_serializer(DateTimeSerializer(), 'T')   # Time
    serialization.register_serializer(DateSerializer(), 'D')     # Date
    serialization.register_serializer(PeriodSerializer(), 'P')   # Period
    serialization.register_serializer(DurationSerializer(), 'I')   # Interval
    serialization.register_serializer(WeekdaySerializer(), 'W')  # Wkday
    serialization.register_serializer(MaskSerializer(), 'M')    # Mask

    db = TinyDB(dbfile, storage=serialization, indent=1, ensure_ascii=False)
    logger.debug(f'db._storage: {db.__dict__}')

    db.default_table_name = 'items'

    # db.insert({'itemtype': '!', 'summary': 'inserted by data.py'})

    return db


def format_duration(obj):
    """
    For etm, microseconds will be zero and seconds will always be an integer
    multiple of 60. Thus timedeltas can always be expressed using only weeks, days,
    hours and minutes.
    """
    if not isinstance(obj, timedelta):
        raise ValueError(f'{obj} is not a timedelta instance')
    until = []
    weeks = obj.days // 7
    days = obj.days % 7
    hours = obj.seconds // (60 * 60)
    minutes = (obj.seconds // 60) % 60
    if weeks:
        until.append(f'{weeks}w')
    if days:
        until.append(f'{days}d')
    if hours:
        until.append(f'{hours}h')
    if minutes:
        until.append(f'{minutes}m')
    if not until:
        until.append('0m')
    return ''.join(until)


def format_duration_list(obj_lst):
    try:
        return ', '.join([format_duration(x) for x in obj_lst])
    except Exception as e:
        print('format_duration_list', e)
        print(obj_lst)


period_regex = re.compile(r'(([+-]?)(\d+)([wdhms]))+?')
threeday_regex = re.compile(r'(MON|TUE|WED|THU|FRI|SAT|SUN)', re.IGNORECASE)
anniversary_regex = re.compile(r'!(\d{4})!')

period_hsh = dict(
    z=timedelta(seconds=0),
    s=timedelta(seconds=1),
    m=timedelta(minutes=1),
    h=timedelta(hours=1),
    d=timedelta(days=1),
    w=timedelta(weeks=1),
)


def parse_duration(s):
    """\
    Take a period string and return a corresponding timedelta.
    Examples:
        parse_duration('-2w3d4h5m')= Duration(weeks=-2,days=3,hours=4,minutes=5)
        parse_duration('1h30m') = Duration(hours=1, minutes=30)
        parse_duration('-10m') = Duration(minutes=10)
    where:
        w: weeks
        d: days
        h: hours
        m: minutes

    >>> 3*60*60+5*60
    11100
    >>> parse_duration("2d-3h5m")[1]
    Duration(days=1, hours=21, minutes=5)
    >>> datetime(2015, 10, 15, 9, 0, tz='local') + parse_duration("-25m")[1]
    DateTime(2015, 10, 15, 8, 35, 0, tzinfo=Timezone('America/New_York'))
    >>> datetime(2015, 10, 15, 9, 0) + parse_duration("1d")[1]
    DateTime(2015, 10, 16, 9, 0, 0, tzinfo=Timezone('UTC'))
    >>> datetime(2015, 10, 15, 9, 0) + parse_duration("1w-2d+3h")[1]
    DateTime(2015, 10, 20, 12, 0, 0, tzinfo=Timezone('UTC'))
    """
    td = period_hsh['z']

    m = period_regex.findall(s)
    if not m:
        return False, "Invalid period '{0}'".format(s)
    for g in m:
        num = -int(g[2]) if g[1] == '-' else int(g[2])
        td += num * period_hsh[g[3]]
    return True, td
